%
% Exports a program specification to a file
%
function exportDefinition(filename, states, transitions, labels, posEx, negEx, extraStates)

    if ~exist('extraStates','var')
        extraStates = 0;
    end

    fd = fopen(filename,"w");
    printPreamble(fd);
    printStateFns(fd, states, extraStates);
    printTransitionFns(fd, states, transitions, labels);
    printExamples(fd, posEx, negEx);
    printPostamble(fd);
    fclose(fd);
end

function printPreamble(fd)
    fprintf(fd, ";;;\n;;; Preamble\n;;;\n");
    fprintf(fd, "(set-logic ALL)\n\n");
end

function printStateFns(fd, states, extraStates)
    fprintf(fd, ";;;\n;;; State Definitions\n;;;\n");

    fprintf(fd, "(define-fun is-state ((q Int)) Bool\n");
    fprintf(fd, "    (and (< (- 1) q) (< q %u)))\n\n", size(states,1)+1+extraStates);

    fprintf(fd, "(define-fun is-accepting ((q Int)) Bool\n");
    fprintf(fd, "    (or");
    any = false;
    for ix=1:size(states,1)
        if states{ix,2}
            fprintf(fd, " (= q %u)", ix);
            any = true;
        end
    end
    if ~any
        fprintf(fd, " false");
    end

    fprintf(fd, "))\n\n");

    fprintf(fd, "(declare-fun IS () Int)\n");
    fprintf(fd, "(assert (is-state IS))\n");
    fprintf(fd, "(assert (not (= 0 IS)))\n\n");
end

function printTransitionFns(fd, states, transitions, labels)
    fprintf(fd, ";;;\n;;; Transition Definitions\n;;;\n");

    fprintf(fd, "(declare-fun T (Int String) Int)\n");

    fprintf(fd, "\n;; All transitions must go to a valid state\n");
    fprintf(fd, "(assert (forall ((q Int) (symb String))\n");
    fprintf(fd, "    (=> (is-state q) (is-state (T q symb)))))\n");

    fprintf(fd, "\n;; It must not be possible to leave the zero state\n");
    fprintf(fd, "(assert (forall ((symb String)) (= 0 (T 0 symb))))\n");
    
    fprintf(fd, "\n;; Defined transitions\n");
    for ix=1:size(transitions,1)
        fprintf(fd, "(assert (= %u (T %u ""%s"")))\n", transitions(ix,2), transitions(ix,1), labels(ix));
    end

    fprintf(fd, "\n;; Negative transitions\n");
    for sIx=1:size(states,1)
        fprintf(fd, "(assert (forall ((symb String))\n");
        fprintf(fd, "    (=> (not (or");
        any = false;
        for tIx=1:size(transitions,1)
            if transitions(tIx,1) == sIx
                fprintf(fd, " (= symb ""%s"")", labels(tIx));
                any = true;
            end
        end
        if ~any
            fprintf(fd, " false");
        end
        fprintf(fd, "))\n");
        fprintf(fd, "        (= 0 (T %u symb)))))\n", sIx);
    end

end

function printExamples(fd, posEx, negEx)
    fprintf(fd, ";;;\n;;; Examples\n;;;\n");

    fprintf(fd, "(define-fun-rec exec-dfa ((q Int) (input String)) Bool\n");
    fprintf(fd, "    (ite (= 0 (str.len input))\n");
    fprintf(fd, "         (is-accepting q)\n");
    fprintf(fd, "         (exec-dfa (T q (str.at input 0)) (str.substr input 1 (- (str.len input) 1)))))\n");

    fprintf(fd, "\n;; Positive\n");
    for ix=1:size(posEx,1)
        fprintf(fd, "(assert (exec-dfa IS ""%s""))\n", posEx(ix));
    end

    fprintf(fd, "\n;; Negative\n");
    for ix=1:size(negEx,1)
        fprintf(fd, "(assert (not (exec-dfa IS ""%s"")))\n", negEx(ix));
    end
end

function printPostamble(fd)
    fprintf(fd, ";;;\n;;; Postamble\n;;;\n");

    fprintf(fd, "(check-sat)\n");
    fprintf(fd, "(get-model)\n");
end