%
% Extracts example boxes from an automaton image
%
% Outputs:
%   positiveEx - string array of positive examples
%   good       - regions for each positive example box
%   negativeEx - string array of negative examples
%   bad        - regions for each negative example box
%   mask       - mask of all pixels used in example boxes
%
function [positiveEx, good, negativeEx, bad, mask] = getBoxes(img)

    red_img = getRedImg(img);
    bad = extractBoxes(red_img);
    green_img = getGreenImg(img);
    good = extractBoxes(green_img);

    positiveEx = strings(size(good,1));
    negativeEx = strings(size(bad,1));

    mask = false(size(img,1:2));

    for ix=1:size(good,1)
        positiveEx(ix) = extractText(img, good(ix));
        mask(good(ix).PixelIdxList) = 1;
    end

    for ix=1:size(bad,1)
        negativeEx(ix) = extractText(img, bad(ix));
        mask(bad(ix).PixelIdxList) = 1;
    end
end

function text = extractText(img, box)
    bw = imbinarize(255 - rgb2gray(img), 'adaptive');
    mask = ones(size(bw),'logical');
    mask(box.PixelIdxList) = 0;
    bw(mask) = 0;

    blur = imbinarize(imgaussfilt(double(bw),20));
    bw(~blur) = 0;
    blur = imbinarize(imgaussfilt(double(bw),10));
    bw(~blur) = 0;
    blur = imbinarize(imgaussfilt(double(bw),5));
    bw(~blur) = 0;

    bw = imclose(bw,strel('disk',2));

    text = "";
    regions = regionprops(bw, 'Image','BoundingBox');
    doOCR = true(size(regions,1));
    for ix=1:size(regions,1)
        if regions(ix).BoundingBox(3) > 1.5*regions(ix).BoundingBox(4)
            doOCR(ix) = false; % Not likely a character due to shape
        end

        if regions(ix).BoundingBox(4) < 30
            doOCR(ix) = false; % Not likely a character due to being too short
        end
    end

    leftmostPosition = Inf;
    leftmost = 1;
    for ix=1:size(regions,1)
        if doOCR(ix)
            if regions(ix).BoundingBox(1) < leftmostPosition
                leftmostPosition = regions(ix).BoundingBox(1);
                leftmost = ix;
            end
        end
    end
    doOCR(leftmost) = false;

    for ix=1:size(regions,1)
        if doOCR(ix)
            text = text + doOcr(img, regions(ix));
%             %ocrResult = ocr(bwmorph(regions(ix).Image,'thin'),'Language','dlang/tessdata/dlang.traineddata','TextLayout','Character');
%             ocrResult = ocr(img,regions(ix).BoundingBox,'Language','dlang/tessdata/dlang.traineddata','TextLayout','Character','CharacterSet','01');
%             if (ocrResult.WordConfidences > 0.2)
%                 % Hack of the century - please forgive me
%                 if ocrResult.Text == '1' && regions(ix).BoundingBox(3) > 30 && ocrResult.WordConfidences < 0.5
%                     text = text + 'B';
%                 elseif ocrResult.Text == 'B' && (regions(ix).BoundingBox(3)/regions(ix).BoundingBox(4)) < 0.3 && ocrResult.WordConfidences < 0.7
%                     text = text + '1'; 
%                 else
%                     text = text + ocrResult.Text;
%                 end
%             end
        end
    end
end

function boxes = extractBoxes(bw)
    blur = imbinarize(imgaussfilt(double(bw),20));
    bw(~blur) = 0;
    blur = imbinarize(imgaussfilt(double(bw),10));
    bw(~blur) = 0;
    blur = imbinarize(imgaussfilt(double(bw),5));
    bw(~blur) = 0;

    props = regionprops(bw,'BoundingBox','Area','Centroid','MajorAxisLength','MinorAxisLength','PixelIdxList');
    areas = zeros(size(props,1));
    for pIx=1:size(props,1)
        % Does it look rectangular?
        p = props(pIx);
        bbArea = p.MajorAxisLength * p.MinorAxisLength;
        if (p.Area > (size(bw,1)*size(bw,2)*0.005)) && (abs(1 - bbArea/p.Area) < 0.4)
            areas(pIx) = p.Area;
        end
    end

    [hist,edges] = histcounts(areas);
    threshold = 0.3;
    limit = edges(end)*threshold;
    boxes = props(areas>limit);
end

function img_red = getRedImg(img)
    img_red = imbinarize(max(img(:,:,1) - rgb2gray(img), 0),'adaptive','Sensitivity',0.1);
    img_red = imfill(img_red,4,'holes');
end

function img_green = getGreenImg(img)
    %hsv_img = rgb2hsv(img);
    %img_green = hsv_img(:,:,1) > (40/255) & hsv_img(:,:,1) < (160/255);
    img_green = imbinarize(max((img(:,:,2)*0.99 + img(:,:,3)*0.01) - rgb2gray(img), 0),'adaptive','Sensitivity',0.8);
    img_green = imfill(img_green,4,'holes');
end

%
% imshow(hsv_img(:,:,1) < (20/255) | hsv_img(:,:,1) > (240/255));
% imshow((hsv_img(:,:,1) < (20/255) | hsv_img(:,:,1) > (240/255)) & hsv_img(:,:,2) > 0.7 & hsv_img(:,:,3) > 0.4);