%
% Extracts states from an automaton image
%
% Outputs:
%  states    - Nx2 cell array. states{:,1} is the region, states{:,2} if accepting
%  stateMask - mask of all pixels associated with a state
function [states, stateMask] = getStates(img)

    mask = preprocess(img);
    stateMask = false(size(mask));

    props = regionprops(mask,'BoundingBox','Centroid','MajorAxisLength','MinorAxisLength','Orientation','FilledImage','FilledArea','Perimeter','Eccentricity','PixelIdxList','ConvexHull');
    candidates = {};

    for regIx=1:size(props)
        [circ, radius] = computeCircularity(props(regIx));
        if (circ > 0.9)
            next = size(candidates,1) + 1;
            candidates{next,1} = props(regIx);
            candidates{next,2} = radius;
        end
    end

    %
    % Filter by radius of circle
    %
    radii = zeros(size(candidates,1));
    for candIx=1:size(candidates,1)
        radii(candIx) = candidates{candIx,2};
    end
    [hist,edges] = histcounts(radii);
    threshold = 0.5; %otsuthresh(hist);
    limit = edges(end)*threshold;

    pstates = cell(0); %nnz(radii>=limit));
    for candIx=1:size(candidates,1)
        if (radii(candIx) >= limit)
            propss = candidates{candIx,1};
            pstates{end+1} = propss;
            stateMask(propss.PixelIdxList) = 1;
        end
    end
    stateMask = imfill(stateMask,4,'holes');

    %
    % Find accepting states
    %
    positions = [];
    for stateIx=1:size(pstates,2)
        positions = [positions; pstates{stateIx}.BoundingBox];
    end
    
    intersection = rectint(positions,positions);
    states = cell(0);
    for stateIx=1:size(pstates,2)
        irow = intersection(stateIx,:);
        if (nnz(irow)==1)
            % Regular state
            states{end+1,1} = pstates{stateIx};
            states{end,2} = 0;
        elseif (min(irow(irow>0))~=intersection(stateIx,stateIx))
            % Accepting state, and this is the outer-most state
            states{end+1,1} = pstates{stateIx};
            states{end,2} = 1;
        end
    end

    %
    % Show the results
    %
    %imshow(img); hold on;

end

%
% Computes the circularity and radius of a region
%
function [circ, radius] = computeCircularity(props)
    props = regionprops(props.FilledImage,'Circularity','MajorAxisLength');
    circ = props.Circularity;
    radius = props.MajorAxisLength / 2;
end

%
% Plots the major and minor axes of a region
%
function [majA minA] = plotAxes(props)
    theta = deg2rad(props.Orientation);
    majL = props.MajorAxisLength;
    minL = props.MinorAxisLength;
    cx = props.Centroid(1);
    cy = props.Centroid(2);

    majA = [cx-cos(theta)*majL/2 cy+sin(theta)*majL/2; cx+cos(theta)*majL/2 cy-sin(theta)*majL/2];
    minA = [];

    line(majA(:,1), majA(:,2),'Color',[0 0 1],'LineWidth',2);
end

function mask = preprocess(img)

img = uint8(img);

% Invert so the foreground is light
img_gray = 255 - rgb2gray(img);
% Top hat transform to deal with uneven backgrounds on white boards
%img_gray = imtophat(img_gray,strel('disk',10));
% Convert to binary
bw = imbinarize(img_gray,'adaptive');

% Blur to create a mask for the pixels we care about
blur = imbinarize(imgaussfilt(double(bw),20));
bw(~blur) = 0;
blur = imbinarize(imgaussfilt(double(bw),10));
bw(~blur) = 0;
mask = imbinarize(imgaussfilt(double(bw),5));

end