%
% Extracts transitions from an automaton image
% 
% Inputs:
%   img    - the original image
%   mask   - a mask of what has been processed in the image so far
%   states - an Nx2 cell array. states{:,1} is the region
%
% Outputs:
%   transitions - an Nx2 array, of "from" and "to" state indices per transition
%   labels      - string labels for each transition
%   tRegions    - region struct for each transition arrow
%   tMask       - mask of all pixels associated with transitions
function [transitions, labels, tRegions, tMask] = getTransitions(img, mask, states)

    regions = regionprops(img_preprocess(img, mask), 'Image', 'BoundingBox', 'PixelIdxList', 'Centroid');
    tMask = false(size(img,1),size(img,2));
    tIx = 1;
    centers = zeros(0);
    transitions = zeros(0);
    for ix=1:size(regions,1)
        [yes, from, to] = processTransition(img, regions(ix).Image, floor(regions(ix).BoundingBox(1:2)), states);
        if yes
            transitions(tIx,1) = from;
            transitions(tIx,2) = to;
            centers(tIx,1) = regions(ix).Centroid(1);
            centers(tIx,2) = regions(ix).Centroid(2);
            tMask(regions(ix).PixelIdxList) = 1;
            tRegions(tIx) = regions(ix);
            tIx = tIx + 1;
        end
    end

    if (size(transitions,1) == 0)
        tRegions = struct;
    end

    labels = strings(size(transitions,1),1);
    labelRegions = regionprops(img_preprocess(img, mask | tMask), 'Image', 'Centroid', 'BoundingBox', 'Area');
    for tIx=1:size(transitions,1)
        mindist = inf;
        label = 0;
        for rIx=1:size(labelRegions,1)
            dist = norm(labelRegions(rIx).Centroid - centers(tIx,:));
            if (dist < mindist) && labelRegions(rIx).Area > 10
                mindist = dist;
                label = rIx;
            end
        end

        if label > 0
            labels(tIx) = doOcr(img, labelRegions(label));
%             ocrResult = ocr(bwmorph(labelRegions(label).Image, 'thin'), ...
%                             'Language','dlang/tessdata/dlang.traineddata', ...
%                             'TextLayout','Character');
%             % Hack of the century - please forgive me
%             if ocrResult.Text == '1' && labelRegions(label).BoundingBox(3) > 30 && ocrResult.WordConfidences < 0.5
%                 labels(tIx) = 'B';
%             elseif ocrResult.Text == 'B' && (labelRegions(label).BoundingBox(3)/labelRegions(label).BoundingBox(4)) < 0.3 && ocrResult.WordConfidences < 0.7
%                 labels(tIx) = '1';
%             else
%                 labels(tIx) = ocrResult.Text;
%             end
        end
    end
end

%
% Tries to extract a transition from a sub image
%
function [isArrow, sFrom, sTo] = processTransition(img, subImg, offset, states)
    [isArrow, head, tail] = getEndpoints(subImg);
    if ~isArrow
        sFrom = 0;
        sTo = 0;
        return;
    end

    midpoint = computeMidpoint(subImg, head, tail);
    hm = computeMidpoint(subImg, head, midpoint);
    tm = computeMidpoint(subImg, tail, midpoint);

    sTo = getIntersectingState(img, hm + offset, head + offset, states);
    sFrom = getIntersectingState(img, tm + offset, tail + offset, states);

    if sTo == 0 || sFrom == 0
        isArrow = false;
    end
end

%
% Finds the first intersecting state
%
function state = getIntersectingState(img, p1, p2, states)
    delta = p2 - p1;
    t = 1;
    step = 1/norm(p1 - p2);
    p = delta * t + p1;
    while (p(1) > 0 && p(2) > 0 && p(1) < size(img,2) && p(2) < size(img,1))
        pointR = [p(1) p(2) 2 2]; % Rectangle around point we're checking
        for sIx=1:size(states,1)
            %hullx = states{sIx,1}.ConvexHull(:,1);
            %hully = states{sIx,1}.ConvexHull(:,2);
            %px = p(1);
            %py = p(2);
            %if inpolygon(px, py, hullx, hully)
            %    state = sIx;
            %    return;
            %end
            if rectint(pointR, states{sIx,1}.BoundingBox) > 0
                state = sIx;
                return;
            end
        end
        t = t + step;
        op = p;
        p = delta * t + p1;
        % line([op(1) p(1)], [op(2) p(2)]);
    end
    state = 0;
end

%
% Gets the head and tail of an arrow
%
function [isArrow, head, tail] = getEndpoints(subImg)
    %skelImg = bwskel(imfill(bwskel(bwmorph(bwmorph(subImg,'close'),'spur')),'holes'),'MinBranchLength',10);
    if bweuler(subImg) < 1
        for r=1:5
            skelImg = imclose(subImg,strel('disk',r));
            if bweuler(skelImg) >= 1
                break;
            end
        end
    else
        skelImg = subImg;
    end
    skelImg = bwmorph(skelImg,'shrink',1);
    skelImg = bwmorph(skelImg,'thin',1);
    skelImg = bwskel(bwmorph(skelImg,'spur'));
    skelImg = bwskel(skelImg, 'MinBranchLength',5);
    skelImg = bwmorph(skelImg,'spur');

    %skelImg = bwmorph(bwmorph(subImg,'spur'),'thin',100);
    %skelImg = bwmorph(bwskel(imfill(skelImg,'holes'),'MinBranchLength',10),'spur');

    if bweuler(skelImg) < 1
        isArrow = false;
        head = zeros();
        tail = zeros();
        return;
    end

    [endY, endX] = find(bwmorph(skelImg,'Endpoints'));

    if size(endY,1)~=3
        isArrow = false;
        head = zeros();
        tail = zeros();
        return;
    end

    p1 = [endX(1) endY(1)];
    p2 = [endX(2) endY(2)];
    p3 = [endX(3) endY(3)];

    if (~subImg(p1(2), p1(1)) || ~subImg(p2(2), p2(1)) || ~subImg(p3(2), p3(1)))
        isArrow = false;
        head = zeros();
        tail = zeros();
        return;
    end

    % Find two closest points
    n12 = connectedDistance(subImg, p1, p2);
    n23 = connectedDistance(subImg, p2, p3);
    n13 = connectedDistance(subImg, p1, p3);

    if (n12 <= n23 && n12 <= n13)
        head = closestTurnedOn(subImg, (p1 + p2) / 2);
        tail = p3;
    elseif (n23 <= n12 && n23 <= n13)
        head = closestTurnedOn(subImg, (p2 + p3) / 2);
        tail = p1;
    elseif (n13 <= n23 && n13 <= n12)
        head = closestTurnedOn(subImg, (p1 + p3) / 2);
        tail = p2;
    else
        error("No minimum?");
    end
    isArrow = true;
end

%
% Gets the closest true image pixel to the given point
%
function pOut = closestTurnedOn(img, pIn)
    x = false(size(img));
    x(ceil(pIn(2)), ceil(pIn(1))) = 1;
    x = bwdist(x);
    x(img == 0) = inf;
    [~,I] = min(x,[],'all');
    [R,C] = ind2sub(size(img),I);
    pOut(1) = C;
    pOut(2) = R;
end

%
% Computes distance between two points along image pixels
%
function d = connectedDistance(img, p1, p2)
    if ~img(p1(2), p1(1))
        error("Not a turned on point (1)");
    end
    
    if ~img(p2(2), p2(1))
        error("Not a turned on point (2)");
    end

    dd = connectedDistanceTransform(img, p1(1), p1(2));
    d = dd(p2(2), p2(1));
end

%
% Computes a connected distance transform - distance from a point along
% connected pixels
%
function d = connectedDistanceTransform(img, x, y)
    d = inf(size(img));
    dp = zeros(size(img));
    d(y, x) = 0;

    ic = 0;
    while (~isequal(dp,d))
        ic = ic+1;
        dp = d;
        for y=1:size(img,1)
            for x=1:size(img,2)
                if img(y,x) && d(y,x) == inf
                    if x > 1 && d(y,x-1) < inf
                        d(y,x) = d(y,x-1) + 1;
                    end

                    if y > 1 && d(y-1,x) < inf
                      d(y,x) = d(y-1,x) + 1;
                    end

                    if x < size(img,2)-1 && d(y,x+1) < inf
                        d(y,x) = d(y,x+1) + 1;
                    end

                    if y < size(img,1)-1 && d(y+1,x) < inf
                        d(y,x) = d(y+1,x) + 1;
                    end
                end
            end
        end
    end
end

%
% Computes the midpoint of two points, along the image pixels
%
function mp = computeMidpoint(img, p1, p2)

    d1 = connectedDistanceTransform(img, p1(1), p1(2));
    d2 = connectedDistanceTransform(img, p2(1), p2(2));

    dd = abs(d1 - d2);
    [~,I] = min(dd, [],'all','omitnan');
    [R,C] = ind2sub(size(img), I);
    mp(1) = C;
    mp(2) = R;
end

%
% Preprocesses an image for transition extraction
%
function bw = img_preprocess(img, mask)
    % Invert so the foreground is light
    img_gray = 255 - rgb2gray(img);
    % Convert to binary
    bw = imbinarize(img_gray,'adaptive');
    bw(mask) = 0;

    % Blur to create a mask for the pixels we care about
    blur = imbinarize(imgaussfilt(double(bw),20));
    bw(~blur) = 0;
    blur = imbinarize(imgaussfilt(double(bw),10));
    bw(~blur) = 0;
    blur = imbinarize(imgaussfilt(double(bw),5));
    bw(~blur) = 0;
    %bw = imbinarize(imgaussfilt(double(bw),1)); % ADD
    %bw = bwmorph(bw, 'close');
    bw = bwmorph(bw,'clean');
end